@@ -1,7 +1,7 @@
 /*****************************************************************************
-*   $Id: main.c,v 7.1 1998/12/07 01:38:00 darren Exp $
+*   $Id: main.c,v 8.20 1999/09/17 07:34:00 darren Exp $
 *
-*   Copyright (c) 1996-1998, Darren Hiebert
+*   Copyright (c) 1996-1999, Darren Hiebert
 *
 *   Author: Darren Hiebert <darren@hiebert.com>, <darren@hiwaay.net>
 *           http://darren.hiebert.com
@@ -23,11 +23,18 @@
 #include "general.h"
 
 #ifdef HAVE_STDLIB_H
-# include <stdlib.h>	    /* to declare malloc(), realloc() */
+# include <stdlib.h>		/* to declare malloc(), realloc() */
 #endif
 #include <string.h>
 #include <errno.h>
 
+#ifndef HAVE_FGETPOS
+# include <stdio.h>		/* to declare SEEK_SET (hopefully) */
+# if defined(HAVE_UNISTD_H)
+#  include <unistd.h>		/* to declare SEEK_SET on SunOS 4.1.x */
+# endif
+#endif
+
 #ifdef AMIGA
 # include <dos/dosasl.h>	/* for struct AnchorPath */
 # include <clib/dos_protos.h>	/* function prototypes */
@@ -65,6 +72,9 @@
 #ifdef HAVE_DIR_H
 # include <dir.h>	/* to declare findfirst() and findnext() */
 #endif
+#ifdef HAVE_IO_H
+# include <io.h>	/* to declare _finddata_t in MSVC++ 4.x */
+#endif
 
 /*  To provide timings features if available.
  */
@@ -83,11 +93,11 @@
 #include "ctags.h"
 #include "debug.h"
 #include "entry.h"
+#include "keyword.h"
 #include "main.h"
 #include "options.h"
 #include "parse.h"
 #include "read.h"
-#include "sort.h"
 #include "vstring.h"
 
 /*============================================================================
@@ -105,9 +115,13 @@
  *  Portability defines
  *--------------------------------------------------------------------------*/
 #if defined(MSDOS) || defined(WIN32) || defined(OS2)
-# define PATH_SEPARATOR	'\\'
+# define PATH_SEPARATOR '\\'
 #else
-# define PATH_SEPARATOR	'/'
+# ifdef QDOS
+#  define PATH_SEPARATOR '_'
+# else
+#  define PATH_SEPARATOR '/'
+# endif
 #endif
 
 /*  File type tests.
@@ -162,21 +176,28 @@ static struct { long files, lines, bytes; } Totals = { 0, 0, 0 };
 #ifdef NEED_PROTO_STAT
 extern int stat __ARGS((const char *, struct stat *));
 #endif
+#ifdef NEED_PROTO_LSTAT
+extern int lstat __ARGS((const char *, struct stat *));
+#endif
 
 # if defined(MSDOS) || defined(WIN32) || defined(AMIGA)
 static boolean createTagsForMatchingEntries __ARGS((char *const pattern));
 #endif
 
+static boolean isSymbolicLink __ARGS((const char *const name));
+static boolean isNormalFile __ARGS((const char *const name));
+static boolean isDirectory __ARGS((const char *const name));
 static boolean createTagsForFile __ARGS((const char *const filePath, const langType language, const unsigned int passCount));
-static vString *getNextListFile __ARGS((FILE *const fp));
 static vString *sourceFilePath __ARGS((const char *const file));
 static boolean createTagsWithFallback __ARGS((const char *const fileName, const langType language));
 static boolean createTagsForDirectory __ARGS((const char *const dirName));
 static boolean createTagsForEntry __ARGS((const char *const entryName));
-static boolean createTagsForList __ARGS((const char *const listFile));
-static boolean createTagsForArgs __ARGS((const char *const *const argList));
+static boolean createTagsFromFileInput __ARGS((FILE* const fp, const boolean filter));
+static boolean createTagsFromListFile __ARGS((const char* const fileName));
+static boolean createTagsForArgs __ARGS((cookedArgs* const args));
 static void printTotals __ARGS((const clock_t *const timeStamps));
-static void makeTags __ARGS((const char *const *const argList));
+static boolean areFilesAvailable __ARGS((cookedArgs* args));
+static void makeTags __ARGS((cookedArgs* arg));
 static void setExecutableName __ARGS((const char *const path));
 
 /*============================================================================
@@ -273,20 +294,35 @@ extern unsigned long getFileSize( name )
     return size;
 }
 
-extern boolean isNormalFile( name )
+static boolean isSymbolicLink( name )
+    const char *const name;
+{
+#if defined(__vms) || defined(MSDOS) || defined(WIN32) || defined(__EMX__) || defined(AMIGA)
+    return FALSE;
+#else
+    struct stat fileStatus;
+    boolean isLink = FALSE;
+
+    if (lstat(name, &fileStatus) == 0)
+	isLink = (boolean)(S_ISLNK(fileStatus.st_mode));
+
+    return isLink;
+#endif
+}
+
+static boolean isNormalFile( name )
     const char *const name;
 {
     struct stat fileStatus;
     boolean isNormal = FALSE;
 
     if (stat(name, &fileStatus) == 0)
-	isNormal = (boolean)(S_ISLNK(fileStatus.st_mode) ||
-			     S_ISREG(fileStatus.st_mode)  );
+	isNormal = (boolean)(S_ISREG(fileStatus.st_mode));
 
     return isNormal;
 }
 
-extern boolean isDirectory( name )
+static boolean isDirectory( name )
     const char *const name;
 {
     boolean isDir = FALSE;
@@ -357,6 +393,27 @@ extern void addTotals( files, lines, bytes )
     Totals.bytes += bytes;
 }
 
+extern boolean isDestinationStdout()
+{
+    boolean toStdout = FALSE;
+
+    if (Option.xref  ||  Option.filter  ||
+	(Option.tagFileName != NULL  &&  (strcmp(Option.tagFileName, "-") == 0
+#if defined(MSDOS) || defined(WIN32) || defined(OS2)
+	/* nothing else special needed here */
+#else
+# ifdef __vms
+    || strncasecmp(Option.tagFileName, "sys$output", sizeof("sys$output")) == 0
+# else
+    || strcmp(Option.tagFileName, "/dev/stdout") == 0
+# endif
+#endif
+	)))
+	toStdout = TRUE;
+
+    return toStdout;
+}
+
 /*----------------------------------------------------------------------------
  *  Pathname manipulation (O/S dependent!!!)
  *--------------------------------------------------------------------------*/
@@ -465,7 +522,7 @@ static boolean createTagsForFile( filePath, language, passCount )
     const langType language;
     const unsigned int passCount;
 {
-    boolean retry = TRUE;
+    boolean retry = FALSE;
 
     if (Option.etags)
 	beginEtagsFile();
@@ -502,29 +559,6 @@ static boolean createTagsForFile( filePath, language, passCount )
     return retry;
 }
 
-static vString *getNextListFile( fp )
-    FILE *const fp;
-{
-    vString *fileName = NULL;
-    int c;
-
-    do
-    {
-	c = fgetc(fp);
-	while (c != '\n'  &&  c != EOF)
-	{
-	    if (fileName == NULL)
-		fileName = vStringNew();
-	    vStringPut(fileName, c);
-	    c = fgetc(fp);
-	}
-	if (fileName != NULL)
-	    vStringTerminate(fileName);
-    } while (fileName == NULL  &&  c != EOF);
-
-    return fileName;
-}
-
 static vString *sourceFilePath( file )
     const char *const file;
 {
@@ -568,8 +602,8 @@ static boolean createTagsForMatchingEntries( pattern )
     char *const pattern;
 {
     boolean resize = FALSE;
-    const size_t tailIndex = baseFilename(pattern) - (char *)pattern;
-    char *const tail = pattern + tailIndex;
+    const size_t dirLength = baseFilename(pattern) - (char *)pattern;
+    vString *const filePath = vStringNew();
 #ifdef HAVE_FINDFIRST
     struct ffblk fileInfo;
     int result = findfirst(pattern, &fileInfo, FA_DIREC);
@@ -582,8 +616,9 @@ static boolean createTagsForMatchingEntries( pattern )
 	 */
 	if (strcmp(entryName, ".") != 0  &&  strcmp(entryName, "..") != 0)
 	{
-	    strcpy(tail, entryName);
-	    resize |= createTagsForEntry(pattern);
+	    vStringNCopyS(filePath, pattern, dirLength);
+	    vStringCatS(filePath, entryName);
+	    resize |= createTagsForEntry(vStringValue(filePath));
 	}
 	result = findnext(&fileInfo);
     }
@@ -602,14 +637,17 @@ static boolean createTagsForMatchingEntries( pattern )
 	     */
 	    if (strcmp(entryName, ".") != 0  &&  strcmp(entryName, "..") != 0)
 	    {
-		strcpy(tail, entryName);
-		resize |= createTagsForEntry(pattern);
+		vStringNCopyS(filePath, pattern, dirLength);
+		vStringCatS(filePath, entryName);
+		resize |= createTagsForEntry(vStringValue(filePath));
 	    }
 	} while (_findnext(hFile, &fileInfo) == 0);
 	_findclose(hFile);
     }
-# endif		/* HAVE__FINDFIRST */
-#endif		/* HAVE_FINDFIRST */
+# endif	/* HAVE__FINDFIRST */
+#endif	/* HAVE_FINDFIRST */
+
+    vStringDelete(filePath);
     return resize;
 }
 
@@ -714,8 +752,13 @@ static boolean createTagsForEntry( entryName )
     boolean resize = FALSE;
     langType language;
 
-    if (! doesFileExist(entryName))
-	error(WARNING | PERROR, "cannot open \"%s\"", entryName);
+    if (isSymbolicLink(entryName)  &&  ! Option.followLinks)
+    {
+	if (Option.verbose)
+	    printf("ignoring %s (symbolic link)\n", entryName);
+    }
+    else if (! doesFileExist(entryName))
+	error(WARNING | PERROR, "cannot access \"%s\"", entryName);
     else if (isDirectory(entryName))
     {
 	if (Option.recurse)
@@ -723,42 +766,43 @@ static boolean createTagsForEntry( entryName )
 	else
 	{
 	    if (Option.verbose)
-		printf("  no recurse into directory %s\n", entryName);
+		printf("ignoring %s (directory)\n", entryName);
 	    resize = FALSE;
 	}
     }
     else if (! isNormalFile(entryName))
     {
 	if (Option.verbose)
-	    printf("  ignoring %s (special file)\n", entryName);
+	    printf("ignoring %s (special file)\n", entryName);
     }
     else if ((language = getFileLanguage(entryName)) == LANG_IGNORE)
     {
 	if (Option.verbose)
-	    printf("  ignoring %s (unknown extension)\n", entryName);
+	    printf("ignoring %s (unknown extension)\n", entryName);
     }
     else
     {
+	if (Option.filter) openTagFile();
 	resize = createTagsWithFallback(entryName, language);
+	if (Option.filter) closeTagFile(resize);
 	addTotals(1, 0L, 0L);
     }
     return resize;
 }
 
-static boolean createTagsForArgs( argList )
-    const char *const *const argList;
+static boolean createTagsForArgs( args )
+    cookedArgs* const args;
 {
     boolean resize = FALSE;
-    const char *const *pArg;
 
     /*  Generate tags for each argument on the command line.
      */
-    for (pArg = argList  ;  *pArg != NULL  ;  ++pArg)
+    while (! cArgOff(args))
     {
-	const char *arg = *pArg;
+	const char *arg = cArgItem(args);
 
 #if defined(MSDOS) || defined(WIN32)
-	vString *const pattern = vStringNew();
+	vString *const pattern = sourceFilePath(arg);
 	char *patternS = vStringValue(pattern);
 
 	vStringCopyS(pattern, arg);
@@ -778,36 +822,59 @@ static boolean createTagsForArgs( argList )
 	resize |= createTagsForEntry(vStringValue(argPath));
 	vStringDelete(argPath);
 #endif
+	cArgForth(args);
+	parseOptions(args);
     }
-
     return resize;
 }
 
-/*  Create tags for the source files listed in the file specified by
- *  "listFile".
+/*  Read from an opened file a list of file names for which to generate tags.
  */
-static boolean createTagsForList( listFile )
-    const char *const listFile;
+static boolean createTagsFromFileInput( fp, filter )
+    FILE* const fp;
+    const boolean filter;
 {
     boolean resize = FALSE;
-    FILE *const fp = (strcmp(listFile,"-") == 0) ? stdin : fopen(listFile, "r");
+    if (fp != NULL)
+    {
+	cookedArgs* args = cArgNewFromFile(fp);
+	parseOptions(args);
+	while (! cArgOff(args))
+	{
+	    const char* const fileName = cArgItem(args);
+	    vString *const filePath = sourceFilePath(fileName);
+	    resize |= createTagsForEntry(vStringValue(filePath));
+	    if (filter)
+	    {
+		if (Option.filterTerminator != NULL)
+		    fputs(Option.filterTerminator, stdout);
+		fflush(stdout);
+	    }
+	    vStringDelete(filePath);
+	    cArgForth(args);
+	    parseOptions(args);
+	}
+	cArgDelete(args);
+    }
+    return resize;
+}
 
-    if (fp == NULL)
-	error(FATAL | PERROR, "cannot open \"%s\"", listFile);
+/*  Read from a named file a list of file names for which to generate tags.
+ */
+static boolean createTagsFromListFile( fileName )
+    const char* const fileName;
+{
+    boolean resize;
+    Assert (fileName != NULL);
+    if (strcmp(fileName, "-") == 0)
+	resize = createTagsFromFileInput(stdin, FALSE);
     else
     {
-	vString *fileName;
-
-	while ((fileName = getNextListFile(fp)) != NULL)
-	{
-	    vString *const filePath = sourceFilePath(vStringValue(fileName));
-
-	    resize |= createTagsForEntry(vStringValue(filePath));
-	    vStringDelete(fileName);
-	    vStringDelete(filePath);
-	}
-	if (fp != stdin)
-	    fclose(fp);
+	FILE* const fp = fopen(fileName, "r");
+	if (fp == NULL)
+	    error(FATAL | PERROR, "cannot open \"%s\"", fileName);
+	resize = createTagsFromFileInput(fp, FALSE);
+	fclose(fp);
     }
     return resize;
 }
@@ -868,7 +935,7 @@ static void printTotals( timeStamps )
 	fprintf(errout, "%lu tag%s sorted", totalTags, plural(totalTags));
 #ifdef CLOCK_AVAILABLE
 	fprintf(errout, " in %.02f seconds",
-		((double)(timeStamps[3] - timeStamps[2])) / CLOCKS_PER_SEC);
+		((double)(timeStamps[2] - timeStamps[1])) / CLOCKS_PER_SEC);
 #endif
 	fputc('\n', errout);
     }
@@ -879,59 +946,68 @@ static void printTotals( timeStamps )
 #endif
 }
 
-static void makeTags( argList )
-    const char *const *const argList;
+static boolean areFilesAvailable( args )
+    cookedArgs* args;
+{
+    return (boolean)(Option.filter || Option.fileList != NULL || !cArgOff(args));
+}
+
+static void makeTags( args )
+    cookedArgs* args;
 {
-    boolean toStdout = FALSE;
     boolean resize = FALSE;
-    clock_t timeStamps[4];
+    clock_t timeStamps[3];
 
-    if (Option.xref  ||  strcmp(Option.tagFileName, "-") == 0
-#if defined(MSDOS) || defined(WIN32) || defined(OS2)
-	/* nothing else special needed here */
-#else
-# ifdef __vms
-    || strncasecmp(Option.tagFileName, "sys$output", sizeof("sys$output")) == 0
-# else
-    || strcmp(Option.tagFileName, "/dev/stdout") == 0
-# endif
-#endif
-	)
-	toStdout = TRUE;
+    if (! areFilesAvailable(args))
+    {
+	if (Option.recurse)
+	{
+	    Assert(cArgOff(args));
+	    cArgDelete(args);
+	    args = cArgNewFromString(".");
+	}
+	else
+	    error(FATAL, "No files specified. Try \"%s --help\".",
+		getExecutableName());
+    }
+#define timeStamp(n) timeStamps[(n)] = (Option.printTotals ? clock():(clock_t)0)
 
-    openTagFile(toStdout);
+    if (! Option.filter)
+	openTagFile();
 
-    if (Option.printTotals) timeStamps[0] = clock();
+    /******/  timeStamp(0);
+
+    if (! cArgOff(args))
+    {
+	if (Option.verbose)
+	    printf("Reading command line arguments\n");
+	resize = createTagsForArgs(args);
+    }
 
     if (Option.fileList != NULL)
-	resize = createTagsForList(Option.fileList);
-    resize = (boolean)(createTagsForArgs(argList) || resize);
-
-    if (Option.printTotals) timeStamps[1] = clock();
-
-    closeTagFile(resize);
-    freeIgnoreList();
-
-    if (TagFile.numTags.added + TagFile.numTags.prev > 0L)
     {
-	if (Option.sorted)
-	{
-	    if (Option.printTotals) timeStamps[2] = clock();
-#ifdef EXTERNAL_SORT
-	    externalSortTags(toStdout);
-#else
-	    internalSortTags(toStdout);
-#endif
-	    if (Option.printTotals) timeStamps[3] = clock();
-	}
-	else if (toStdout)
-	    catFile(tagFileName());
+	if (Option.verbose)
+	    printf("Reading list file\n");
+	resize = (boolean)(createTagsFromListFile(Option.fileList) || resize);
     }
-    if (toStdout)
-	remove(tagFileName());		/* remove temporary file */
+
+    if (Option.filter)
+    {
+	if (Option.verbose)
+	    printf("Reading filter input");
+	resize = (boolean)(createTagsFromFileInput(stdin, TRUE) || resize);
+    }
+
+    /******/  timeStamp(1);
+
+    if (! Option.filter)
+	closeTagFile(resize);
+
+    /******/  timeStamp(2);
 
     if (Option.printTotals)
 	printTotals(timeStamps);
+#undef timeStamp
 }
 
 /*----------------------------------------------------------------------------
@@ -953,41 +1029,39 @@ extern int main( argc, argv )
     int __unused__ argc;
     char **argv;
 {
-    char **envArgList;
-    const char *const *fileList;
+    cookedArgs *args;
+
+#ifdef AMIGA
+    /* This program doesn't work when started from the Workbench */
+    if (argc == 0)
+	exit(1);
+#endif
 
 #ifdef __EMX__
-    _wildcard(&argc, &argv);		/* expand wildcards in argument list */
+    _wildcard(&argc, &argv);	/* expand wildcards in argument list */
 #endif
 
     setExecutableName(*argv++);
     testEtagsInvocation();
 
-    /*  Parse options.
-     */
-    envArgList = parseEnvironmentOptions();
-    fileList = (const char *const *)parseOptions(argv);
-    setOptionDefaults();
+    args = cArgNewFromArgv(argv);
+    previewFirstOption(args);
+    initOptions();
+    readOptionConfiguration();
+    if (Option.verbose)
+	printf("Reading options from command line\n");
+    parseOptions(args);
+    checkOptions();
+    makeTags(args);
 
-    /*	Generate tags if there is at least one source file or a file list.
+    /*  Clean up.
      */
-    if (*fileList == NULL  &&  Option.fileList == NULL)
-    {
-	if (! Option.recurse)
-	    error(FATAL, "No files specified. Try \"%s --help\".",
-		getExecutableName());
-	else
-	{
-	    static const char *defaultRecursionDir[] = { ".", NULL };
-	    fileList = defaultRecursionDir;
-	}
-    }
-    makeTags(fileList);
+    cArgDelete(args);
+    freeKeywordTable();
+    freeSourceFileResources();
+    freeTagFileResources();
+    freeOptionResources();
 
-    /*  Post tag generation clean-up.
-     */
-    if (envArgList != NULL)
-	free(envArgList);
     exit(0);
     return 0;
 }
